/*
 * routines for dealing with config file
 */

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <libgen.h>
#include <string.h>
#include "cf.h"
#include "concat.h"
#include "dict.h"
#include "expr.h"
#include "field.h"
#include "pathnames.h"
#include "talloc.h"

/*
 * config file syntax:
 *
 *
 * [stanza1]
 * /pattern1/ expr
 * /pattern2/ expr
 * name1:def1
 * name2:def2
 * [stanza2]
 * /pattern1/ expr
 * /pattern2/ expr
 * name1:def1
 * name3:def3
 * 
 * should produce
 * 
 * ((stanza1 (:args (pattern1 expr) (pattern2 expr))
 *           (name1 def1) 
 *           (name2 def2)))
 * ((stanza2 (:args (pattern1 expr) (pattern2 expr))
 *           (name1 def1) 
 *           (name3 def2))))
 */

     
/* return 1 iff 's' contains nothing but white space */

static int
blankline (char *s)
{
    while (isspace (*s))
	++s;
    return (*s == '\0');
}

/* change all NL and CR characters to SP, remove trailing SPs */

static void
cleanup (char *s)
{
    char *p = s;

    while (*p) {
	if (*p == '\n' || *p == '\r')
	    *p = ' ';
	++p;
    }

    /* now remove trailing spaces */
    while (p > s && p[-1] == ' ')
	*--p = '\0';
}

/* return the number of NL characters in 's' */

static int
countnl (char *s)
{
    int result = 0;

    while (*s)
	if (*s++ == '\n')
	    ++result;
    return result;
}

/* return a pointer to the first non-space char in 's' */

static char *
firstnonspace (char *s)
{
    while (isspace (*s))
	++s;
    return s;
}

Atom *
get_config (char *filename, FILE *fp, char *compiler)
{
    char buf[10240];
    char *ptr;
    Atom *result = NULL;
    Atom *args = NULL;
    int lineno = 1;
    int copying = 0;

    while (get_header_field (buf, sizeof (buf), fp)) {
	int nl = countnl (buf);

	cleanup (buf);
	if (blankline (buf))
	    ;
	else if (*buf == '[') {			/* [compiler] */
	    if ((ptr = strchr (buf+1, ']')) == NULL) {
		fprintf (stderr, "%s (%d): missing ']'\n",
			 filename, lineno);
	    }
	    else {
		*ptr = '\0';
		copying = (strcmp (buf+1, compiler) == 0);
	    }
	}
	else if (*buf == '/') {			/* /pattern/ expression */
	    if ((ptr = strchr (buf+1, '/')) == NULL) {
		fprintf (stderr, "%s (%d): missing '/' to end pattern\n",
			 filename, lineno);
	    }
	    else if (copying) {
		*ptr = '\0';
		args = define (args, buf+1, string (firstnonspace (ptr+1)));
	    }
	    
	}
	else if (ptr = strchr (buf, ':')) {	/* name: value */
	    if (copying) {
		*ptr = '\0';
		result = define (result, buf,
				 string (firstnonspace (ptr+1)));
	    }
	}
	else {
	    fprintf (stderr, "%s (%d): unrecognized config line\n",
		     filename, lineno);
	    fprintf (stderr, "<%s>\n", buf);
	}

	lineno += nl;
    }
    if (args)
	result = define (result, ":args", args);
    return result;
}

Atom *
find_config (char *filename, char *compiler)
{
    FILE *fp;
    Atom *result;
    char *pathname = NULL;

    /*
     * check in local directory first; if not found, check config directory
     *
     * XXX should allow multiple levels of override:
     * 1. local directory
     * 2. config dir under $HOME/NetBuild2
     * 3. installation or system config dir
     *
     * XXX should also allow per-compiler override leaving default
     * settings for other compilers
     */
    if ((fp = fopen (filename, "r")) != NULL) {
	pathname = strdup (filename);
    }
    else {
	pathname = concat (PATH_CONFIG_DIR, "/", filename, NULL);
	if ((fp = fopen (pathname, "r")) == NULL) {
	    fprintf (stderr, "cannot open %s: %s\n", pathname,
		     strerror (errno));
	    free (pathname);
	    return NULL;
	}
    }

    result = get_config (pathname, fp, compiler);
    if (result == NULL && *compiler == '/') {
	rewind (fp);
	result = get_config (pathname, fp, basename (compiler));
    }
    fclose (fp);
    free (pathname);
    return result;
}

int
match_pattern (char *pat, char *arg)
{
    int plen;
    int arglen;

    if (strcmp (pat, arg) == 0)
	return 1;
    plen = strlen (pat);
    arglen = strlen (arg);
    if (*pat == '*' && (plen - 1 <= arglen) &&
	strcmp (arg + arglen - plen - 1, pat + 1) == 0)
	return 1;
    else if (pat[plen-1] == '*' && strncmp (arg, pat, plen-1) == 0)
	return 1;
    return 0;
}

Atom *
lookup_pattern (Atom *dict, char *name)
{
    if (nullp (dict))
	return NULL;
    else if (!listp (dict))
	return NULL;
    else if (stringp (this (this (dict))) && match_pattern (stringval (this (this (dict))), name))
	return this (dict);
    else
	return lookup_pattern (next (dict), name);
}

Atom *
find_pattern (char *arg, Atom *config)
{
    Atom *result;

    /*
     * try exact match first.  if that fails, look for a matching pattern
     */
    if (result = lookup (config, arg))
	return result;
    return lookup_pattern (config, arg);
}


/************************************************************************/
#if 0

int skip = 0;
Atom *libs = NULL;
Atom *dirs = NULL;
Atom *mode;
int verbose = 0;
Atom *dict = NULL;
char **args;

Atom *
fn_skip (Atom *a)
{
    if (!listp (a) || !nullp (next(a)))
	return error ();
    a = this (a);
    if (!integerp (a) || intval (a) < 0)
	return error ();
    skip = intval (a);
    return integer (0);
}

static int
findlib (Atom *list, Atom *newlib)
{
    if (!listp (list))
	return 0;
    if (!listp (this (list)))
	return 0;
    if (stringp (this (this (list))) &&
	strcmp (stringval (this (this (list))), stringval (newlib)) == 0)
	return 1;
    return findlib (next (list), newlib);
}

Atom *
fn_addlib (Atom *a)
{
    if (!listp (a) || !nullp (next(a)))
	return error ();
    a = this (a);
    if (!stringp (a))
	return error ();
    if (findlib (libs, a))
	;
    else
	libs = append (libs, cons (cons (a, cons (cons (string (":mode"), mode), NULL)), NULL));
    return integer (0);
}

Atom *
fn_adddir (Atom *a)
{
    if (!listp (a) || !nullp (next(a)))
	return error ();
    a = this (a);
    if (!stringp (a))
	return error ();
    dirs = append (dirs, cons (a, NULL));
    return integer (0);
}

Atom *
fn_setmode (Atom *a)
{
    if (!listp (a) || !nullp (next(a)))
	return error ();
    a = this (a);
    if (!stringp (a))
	return error ();
    mode = ref (a);
    return integer (0);
}

Atom *
fn_setverbose (Atom *a)
{
    if (!listp (a) || !nullp (next(a)))
	return error ();
    a = this (a);
    if (!integerp (a))
	return error ();
    verbose = intval (a);
    return integer (0);
}

Atom *
fn_arg (Atom *a)
{
    int offset;
    int i;

    if (!listp (a) || !nullp (next(a)))
	return error ();
    a = this (a);
    if (!integerp (a))
	return error ();
    offset = intval (a);
    if (offset < 0)
	return string ("");
    for (i = 0; i < offset; ++i)
	if (args[i] == NULL)
	    return string ("");
    return string (args[offset]);
}

Atom *
fn_substr (Atom *a)
{
    Atom *str;
    int first;
    int len;

    if (!listp (a) || !listp (next(a)))
	return error ();
    if (!stringp (this (a)))
	return error ();
    str = this (a);
    if (!integerp (this (next (a))))
	return error ();
    first = intval (this (next (a)));
    if (nullp (next (next (a))))
	len = strlen (stringval (str)) - first;
    else if (integerp (next (next (a))))
	len = intval (next (next (a)));
    else
	return error ();
    if (len < 0)
	return string ("");
    return stringn (stringval (str) + first, len);
}


Atom *
lookup_id (Atom *a)
{
    return next (lookup (dict, stringval (a)));
}

void
process_args (char **argv, Atom *root)
{
    Atom *config = lookup (root, ":args");
    int i;

    for (i = 0; argv[i]; ++i) {
	Atom *pat = find_pattern (argv[i], config);

	skip = 0;
	args = argv + i;

	printf ("argv[%d] = %s ", i, argv[i]);
	fprint_atom (stdout, next (pat));
	printf (" => ");
	print_atom (expr_evaluate (next (pat), lookup_id));
	printf ("\n");

	printf ("skip = %d\n", skip);
	printf ("libs = ");
	print_atom (libs);
	printf ("\n");
	printf ("dirs = ");
	print_atom (dirs);
	printf ("\n");
	printf ("mode = %s\n", stringval (mode));
	printf ("verbose = %d\n", verbose);

	i += skip;
    }
}

Atom *
path2list (char *s)
{
    char *colon;

    if (s == NULL)
	return NULL;
    else if (*s == '\0')
	return cons (string (""), NULL);
    else if (colon = strchr (s, ':')) 
	return cons (stringn (s, colon-s), path2list (colon+1));
    else
	return cons (string (s), NULL);
}

main (int argc, char **argv)
{
    Atom *root = find_config ("default.config", argv[1]);
    Atom *argpats = lookup (root, ":args");
    Atom *path;
    int i;

#if 0
    fprint_atom (stdout, root);
    printf ("\n");
#endif
    dict = define (dict, "adddir", function (fn_adddir));
    dict = define (dict, "addlib", function (fn_addlib));
    dict = define (dict, "arg", function (fn_arg));
    dict = define (dict, "setmode", function (fn_setmode));
    dict = define (dict, "setverbose", function (fn_setverbose));
    dict = define (dict, "skip", function (fn_skip));
    dict = define (dict, "substr", function (fn_substr));

    mode = string ("ds");

    if (path = lookup (root, "Library-Path"))
	dirs = path2list (stringval (next (path)));

    process_args (argv+1, root);
}
#endif
