/*
 * routines for getting information about PowerPC CPUs
 */

#ifdef CPU_powerpc

#include <stdio.h>
#include <sys/types.h>

#include "conf.h"
#include "dict.h"
#include "global.h"
#include "type.h"

extern struct global global;

static void
set_value_int (char *name, int value)
{
    global.platform_attrs = 
	define (global.platform_attrs, name, integer (value));
}

static void
set_value_str (char *name, char *value)
{
    global.platform_attrs = 
	define (global.platform_attrs, name, string (value));
}

/*
 * code to test the presence of AltiVec
 * adapted from http://developer.apple.com/hardware/ve/g3_compatibility.html
 */

#include <signal.h>
#include <setjmp.h>

static volatile int has_altivec = -1L;

static sigjmp_buf sigjmp_env;

static void
sig_ill_handler( int sig ) 
{
    /* Set our flag to 0 to indicate AltiVec is illegal */
    has_altivec = 0;

    /* long jump back to safety */
    siglongjmp(sigjmp_env, 0);
}

/* 
 * XXX need to make sure this function is not inlined, else the compiler
 * might add code to the calling function to save/restore AltiVec registers
 * which would cause an illegal instruction before we had the signal
 * handler set up to catch it.
 *
 * maybe place this function in a separate file?
 */

static int
powerpc_try_altivec ( void )
{
#if defined( __GNUC__ )
    asm volatile ( "vor v0, v0, v0" );
#elif defined( __MWERKS__ )
    asm{ vor v0, v0, v0 }
#else
#error unknown compiler
#endif
}

static void
powerpc_test_altivec ( void )
{
    if (has_altivec == -1L) {

	sig_t oldhandler;
	sigset_t signame;
	struct sigaction sa_new, sa_old;

	/* Set AltiVec to ON */
	has_altivec = 1;
	
	/* Set up the signal mask */
	sigemptyset (&signame);
	sigaddset (&signame, SIGILL);

	/* Set up the signal handler */
	sa_new.sa_handler = sig_ill_handler;
	sa_new.sa_mask = signame;
	sa_new.sa_flags = 0;
	sigaction (SIGILL, &sa_new, &sa_old);
	
	/* Attempt to use AltiVec */
	if (sigsetjmp (sigjmp_env, 0) == 0)
	    powerpc_try_altivec ();

	/* Restore the old signal handler */
	sigaction (SIGILL, &sa_old, &sa_new);
    }

    set_value_int ("target.cpu.powerpc.altivec", has_altivec);
}



#ifdef OS_darwin

/*
 * routines for getting hardware info from MacOS X
 *
 * XXX
 * should eventually support a general sysctl interface, so that
 * attributes like target.sysctl.hw.ncpu can be queried.  
 *
 * unfortunately, sysctl is thoroughly borked.  I tried traversing the
 * lists of "mib" style attributes (attributes whose names consist of
 * a list of integers) from <sys/sysctl.h> but that missed some
 * important attributes that don't have mib names, or don't have them
 * in the include file.  And there doesn't seem to be any documented
 * way of scanning all sysctl attributes (other than calling the
 * sysctl program, and that doesn't show type information).
 */

#ifdef HAVE_SYS_SYSCTL_H
#include <sys/sysctl.h>
#endif


static void
darwin_count_cpus ()
{
    size_t oldlen;
    int ncpu;

    oldlen = sizeof (ncpu);
    if (sysctlbyname ("hw.ncpu", &ncpu, &oldlen, NULL, 0) == 0)
	set_value_int ("target.ncpu", ncpu);
}


#include <mach-o/arch.h>

static void
darwin_set_cpu ()
{
    NXArchInfo *foo;

    if ((foo = (NXArchInfo *) NXGetLocalArchInfo ()) !=  NULL)
	set_value_str ("target.cpu.powerpc.model", (char *) foo->name);
}

#endif /* OS_darwin */

#if OS_aix
#include <sys/systemcfg.h>

aix_set_cpu ()
{
    char *foo = NULL;

    /* 
     * AIX returns both 'implementation' (presumably a generic cpu type)
     * and 'version' (presumably a more specific type).   Use 'version'
     * if we know how to decode it, else implementation.  Eventually
     * try to return both for all powerpc processors.
     */
    switch (_system_configuration.implementation) {
        case POWER_RS1: foo="rs1"; break;
        case POWER_RSC: foo="rsc"; break;
        case POWER_RS2: foo="rs2"; break;
        case POWER_601: foo="ppc601"; break;
        case POWER_603: foo="ppc603"; break;
        case POWER_604: foo="ppc604"; break;
        case POWER_620: foo="ppc620"; break;
        case POWER_630: foo="ppc630"; break;
        case POWER_A35: foo="ppca35"; break;
        case POWER_RS64II: foo="rs64-ii"; break;
        case POWER_RS64III: foo="rs64-iii"; break;
        case POWER_RS64IV: foo="power4"; break;
        case POWER_MPC7450: foo="mpc7450"; break;
    }
    switch (_system_configuration.version) {
        case PV_601:    foo="ppc601"; break;
        case PV_601a:   foo="ppc601a"; break;
        case PV_603:    foo="ppc603"; break;
        case PV_604:    foo="ppc604"; break;
        case PV_620:    foo="ppc620"; break;
        case PV_630:    foo="ppc630"; break;
        case PV_A35:    foo="ppca35"; break;
        case PV_RS64II: foo="ppcrs64ii"; break;
        case PV_RS64III: foo="ppcrs64iii"; break;
        case PV_4:      foo="ppc4"; break;
        case PV_MPC7450: foo="ppc7450"; break;
        case PV_4_2:    foo="ppc4_2"; break;
        case PV_RS2:    foo="rs2"; break;
        case PV_RS1:    foo="rs1"; break;
        case PV_RSC:    foo="rsc"; break;
        case PV_M1:     foo="ia64-m1"; break;
        case PV_M2:     foo="ia64-m2"; break;
    }
    if (foo)
    	set_value_str ("target.cpu.powerpc.model", foo);
    set_value_int ("target.ncpu", _system_configuration.ncpus);
    /* 
     * XXX eventually fill in cache sizes 
     */
}
#endif


/*
 * XXX need to at least test for altivec
 */

void
cpu_init ()
{
    set_value_str ("target.cpu.arch", "powerpc");

#ifdef OS_darwin
    darwin_count_cpus ();
    darwin_set_cpu ();
#endif /* OS_darwin */

#ifdef OS_aix
    aix_set_cpu ();
#endif

    powerpc_test_altivec ();
}
#endif /* CPU_powerpc */
