%{
/*
 * evaluator for expressions used in metadata constraints and
 * in user-configurable actions associated with command-line arguments.
 *
 * error reporting: special function return value for expression errors
 * error messages: none
 */

#include "conf.h"

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_STRING_H
#include <string.h>
#else
char *strdup ();
#endif
#include "type.h"
#include "expr.h"

static char yytext[1000];

static Atom *result ;
static int yyerror ();
static Atom *get_id_value ();
%}

%start expression

%token CONST ID STRING 

%left ';'
%right '?' ':'
%left OR
%left AND
%left '|'
%left '^'
%left '&'
%left EQ NEQ
%left '<' '>' LEQ GEQ
%left RSHIFT LSHIFT
%left '+' '-'
%left '*' '/' '%'
%right UMINUS '!' '~'
%union {
    Atom *atom;
}
%type <atom> expression expr id arglist STRING CONST ID

%%
expression: expr
		{
		    result = $1;
		}
	;

id: ID
		{
		    extern char yytext[];

		    $$ = string (yytext);
		}
	;


arglist: expr
		{
		    $$ = cons ($1, NULL);
		}
	| arglist ',' expr
		{
		    $$ = append ($1, cons ($3, NULL));
		}
	;

expr: 
	id
		{
		    $$ = ref (get_id_value ($1, NULL));
		    unref ($1);
		}
	| id '(' arglist ')'
		{
		    $$ = ref (get_id_value ($1, $3));
		    unref ($1);
		    unref ($3);
		}
	| id '(' ')'
		{
		    Atom *emptylist = cons (NULL, NULL);

		    $$ = ref (get_id_value ($1, emptylist));
		    unref ($1);
		    unref (emptylist);
		}
	| CONST
		{
		    extern char yytext[];

		    $$ = integer (strtol(yytext, (char **)NULL, 00));
		}
	| STRING
		{
		    $$ = string (yytext);
		}
	| '(' expr ')'
		{
		    $$ = $2;
		}
	| '-' expr	%prec UMINUS
		{
		    if (integerp ($2))
			$$ = integer (- intval ($2));
		    else
			$$ = error ();
		    unref ($2);
		}
	| '!' expr
		{
		    if (integerp ($2))
			$$ = integer ( ! intval ($2));
		    else
			$$ = error ();
		    unref ($2);
		}
	| '~' expr
		{
		    if (integerp ($2))
			$$ = integer ( ~ intval ($2));
		    else
			$$ = error ();
		    unref ($2);
		}
	| expr '*' expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) * intval ($3));
		    else
			$$ = error ();
		    unref ($1);
		    unref ($3);
		}
	| expr '/' expr
		{
		    if (integerp ($1) && integerp ($3) && intval ($3) != 0)
			$$ = integer (intval ($1) / intval ($3));
		    else
			$$ = error ();
		    unref ($1);
		    unref ($3);
		}
	| expr '%' expr
		{
		    if (integerp ($1) && integerp ($3) && intval ($3) != 0)
			$$ = integer (intval ($1) % intval ($3));
		    else
			$$ = error ();
		    unref ($1);
		    unref ($3);
		}
	| expr '+' expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) + intval ($3));
		    else if (stringp ($1) && stringp ($3))
			$$ = string_concat ($1, $3, NULL);
		    else
			$$ = error ();
		    unref ($1);
		    unref ($3);
		}
	| expr '-' expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) - intval ($3));
		    else
			$$ = error ();
		    unref ($1);
		    unref ($3);
		}
	| expr RSHIFT expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) >> intval ($3));
		    else
			$$ = error ();
		    unref ($1);
		    unref ($3);
		}
	| expr LSHIFT expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) << intval ($3));
		    else
			$$ = error ();
		    unref ($1);
		    unref ($3);
		}
	| expr '<' expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) < intval ($3));
		    else if (stringp ($1) && stringp ($3))
			$$ = integer (strcmp (stringval ($1),
					      stringval ($3)) < 0);
		    else
			$$ = integer (0);
		    unref ($1);
		    unref ($3);
		}
	| expr '>' expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) > intval ($3));
		    else if (stringp ($1) && stringp ($3))
			$$ = integer (strcmp (stringval ($1),
					      stringval ($3)) > 0);
		    else
			$$ = integer (0);
		    unref ($1);
		    unref ($3);
		}
	| expr LEQ expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) <= intval ($3));
		    else if (stringp ($1) && stringp ($3))
			$$ = integer (strcmp (stringval ($1),
					      stringval ($3)) <= 0);
		    else
			$$ = integer (0);
		    unref ($1);
		    unref ($3);
		}
	| expr GEQ expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) >= intval ($3));
		    else if (stringp ($1) && stringp ($3))
			$$ = integer (strcmp (stringval ($1),
					      stringval ($3)) >= 0);
		    else
			$$ = integer (0);
		    unref ($1);
		    unref ($3);
		}
	| expr EQ expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) == intval ($3));
		    else if (stringp ($1) && stringp ($3))
			$$ = integer (strcmp (stringval ($1),
					      stringval ($3)) == 0);
		    else
			$$ = integer (0);
		    unref ($1);
		    unref ($3);
		}
	| expr NEQ expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) != intval ($3));
		    else if (stringp ($1) && stringp ($3))
			$$ = integer (strcmp (stringval ($1),
					      stringval ($3)) != 0);
		    else
			$$ = integer (0);
		    unref ($1);
		    unref ($3);
		}
	| expr '&' expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) & intval ($3));
		    else
			$$ = error ();
		    unref ($1);
		    unref ($3);
		}
	| expr '^' expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) ^ intval ($3));
		    else
			$$ = error ();
		    unref ($1);
		    unref ($3);
		}
	| expr '|' expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) | intval ($3));
		    else
			$$ = error ();
		    unref ($1);
		    unref ($3);
		}
	| expr AND expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) && intval ($3));
		    else
			$$ = error ();
		    unref ($1);
		    unref ($3);
		}
	| expr OR expr
		{
		    if (integerp ($1) && integerp ($3))
			$$ = integer (intval ($1) || intval ($3));
		    else
			$$ = error ();
		    unref ($1);
		    unref ($3);
		}
	| expr '?' expr ':' expr
		{
		    if (!integerp ($1))
			$$ = error ();
		    $$ = intval ($1) ? $3 : $5;
		    unref ($1);
		}
	| expr ';' expr
		{
		    $$ = $3;
		    unref ($1);
		}
	| '{' expr '}'
		{
		    $$ = $2;
		}
	;

%%
#include "global.h"
#include "talloc.h"

extern struct global global;

/* eschew ctype because of Posix locale brain-damage */

#define ISDIGIT(c) ((c) >= '0' && (c) <= '9')
#define ISODIGIT(c) ((c) >= '0' && (c) <= '7')
#define ISXDIGIT(c) (ISDIGIT(c) || ((c) >= 'a' && (c) <= 'f') || ((c) >= 'A' && (c) <= 'F'))
#define ISUPPER(c) ((c) >= 'A' && (c) <= 'Z')
#define ISLOWER(c) ((c) >= 'a' && (c) <= 'z')
#define ISALPHA(c) (ISUPPER(c) || ISLOWER(c))
#define ISALNUM(c) (ISALPHA(c) || ISDIGIT(c))

static char *ptr;
static int errflag = 0;
static Atom *(*idlookup)();
static Atom *idcontext;

struct {
    char *str;
    int sym;
} twos[] = {
    { ">>", RSHIFT },
    { "<<", LSHIFT },
    { "<=", LEQ },
    { ">=", GEQ },
    { "==", EQ },
    { "!=", NEQ },
    { "&&", AND },
    { "||", OR },
};

static char *ones = "()-!~*/%+<>&|^?:,{};";

static int
unhex (c)
char c;
{
    if (c >= 'A' && c <= 'F')
	return c - 'A' + 10;
    if (c >= 'a' && c <= 'f')
	return c - 'a' + 10;
    if (c >= '0' && c <= '9')
	return c - '0';
}

static int
yylex ()
{
    int i;

    while (*ptr == ' ' || *ptr == '\t' || *ptr == '\r' || *ptr == '\n')
	++ptr;

    if (*ptr == '\0')
	return 0;

    /* two-character symbols */
    for (i = 0; i < sizeof(twos)/sizeof(*twos); ++i) {
	if ((ptr[0] == twos[i].str[0]) && (ptr[1] == twos[i].str[1])) {
	    yytext[0] = ptr[0];
	    yytext[1] = ptr[1];
	    yytext[2] = '\0';
	    ptr += 2;
	    return twos[i].sym;
	}
    }
    /* one-character symbols */
    for (i = 0; ones[i]; ++i) {
	if (ptr[0] == ones[i]) {
	    yytext[0] = ones[i];
	    yytext[1] = '\0';
	    ++ptr;
	    return ones[i];
	}
    }
    /* integer constants (hex, octal, decimal) */
    if (ISDIGIT (*ptr)) {
	char *dst = yytext;

	if (*ptr == '0' && (ptr[1] == 'x' || ptr[1] == 'X')) {
	    *dst++ = *ptr++;
	    *dst++ = *ptr++;
	    while (dst < (yytext + sizeof(yytext) - 1) && ISXDIGIT (*ptr))
		*dst++ = *ptr++;
	    *dst++ = '\0';
	}
	else if (*ptr == '0') {
	    *dst++ = *ptr++;
            while (dst < (yytext + sizeof(yytext) - 1) && ISODIGIT (*ptr))
                *dst++ = *ptr++;
            *dst++ = '\0';
	}
	else {
	    while (dst < (yytext + sizeof(yytext) - 1) && ISDIGIT (*ptr))
	        *dst++ = *ptr++;
	    *dst++ = '\0';
	}
	return CONST;
    }
    /* identifiers */
    else if (ISALPHA (*ptr)) {
	char *dst = yytext;

        while (dst < (yytext + sizeof(yytext) - 1) && 
               ISALNUM (*ptr) || *ptr == '.' || *ptr == '_' || *ptr == '$') 
            *dst++ = *ptr++;
        *dst++ = '\0';
	return ID;
    }
    /* quoted strings */
    else if (*ptr == '"') {
	char *dst = yytext;

	++ptr;
        while (dst < (yytext + sizeof(yytext) - 2)) {
	    int octal = 0;

	    if (*ptr == '\0')
		return 0;
	    else if (*ptr == '"') {
		++ptr;
		*dst++ = '\0';
		return STRING;
	    }
	    if (*ptr == '\\') {
		++ptr;
		switch (*ptr) {
		case 'a': *dst++ = '\007'; ptr++; break;
		case 'b': *dst++ = '\b'; ptr++; break;
		case 'e': *dst++ = '\033'; ptr++; break;
		case 'f': *dst++ = '\f'; ptr++; break;
		case 'n': *dst++ = '\n'; ptr++; break;
		case 'r': *dst++ = '\r'; ptr++; break;
		case 't': *dst++ = '\t'; ptr++; break;
		case 'v': *dst++ = '\v'; ptr++; break;
		case '\\': *dst++ = '\\'; ptr++; break;
		case 'x':
		    ++ptr;
		    octal = 0;
		    if (ISXDIGIT (*ptr))
			octal = unhex (*ptr++);
		    if (ISXDIGIT (*ptr))
			octal = octal * 16 + unhex (*ptr++);
		    *dst++ = octal;
		    break;
		default:
		    if (ISDIGIT (*ptr)) {
			octal = *ptr++ - '0'; 
			if (ISDIGIT (*ptr))
			    octal = octal * 8 + *ptr++ - '0';
			if (ISDIGIT (*ptr))
			    octal = octal * 8 + *ptr++ - '0';
			*dst++ = octal;
		    }
		    else
			*dst++ = *ptr++;
		}
	    }
	    else
		*dst++ = *ptr++;
	}
    }
    /* anything else */
    else {
	yytext[0] = *ptr;
	yytext[1] = '\0';
	return *ptr++;
    }
}

static int
yyerror ()
{
    errflag = 1;
}

static Atom *
get_id_value (Atom *name, Atom *arglist)
{
    Atom *val;
    Atom *result;

    if (!stringp (name))
	return error ();
    val = (*idlookup)(stringval (name), idcontext);
    if (nullp (val))
	return integer (0);
    switch (val->type) {
    case integerType:
    case stringType:
    case errorType:
	return val;
    case functionType:
	return (*(((Function *)val)->func))(arglist);
    default:
	return error ();
    }
}

Atom *
expr_evaluate (Atom *s, Atom *(*callback)(), Atom *context)
{
    if (!stringp (s))
	return error ();
    if (callback == NULL)
	return error ();
    idlookup = callback;
    idcontext = context;
    ptr = stringval (s);
    errflag = 0;
    yyparse ();
    if (errflag)
	return error ();
    return result;
}

/* 
 * XXX need to figure out how to implement the callback function for the
 * command-line expression evaluator
 */

nb_expr (char **argv)
{
    int i;
    int len;
    char *expr;
    Atom *result;
    extern Atom *lookup_id ();

    len = 0;
    for (i = 0; argv[i]; ++i)
	len += strlen (argv[i]) + 1;
    expr = TALLOC(len+1, char);
    *expr = '\0';
    for (i = 0; argv[i]; ++i) {
	strcat (expr, argv[i]);
	strcat (expr, " ");
    }
    result = expr_evaluate (string (expr), lookup_id, global.platform_attrs);
    print_atom (result);
    printf ("\n", expr);
}

#ifdef TEST
Atom *
test (Atom *arg)
{
    printf ("test argument:\n");
    print_atom (arg);
    printf ("\n");
    return integer (0);
}

Atom *
test_cb (Atom *name)
{
    if (!stringp (name))
	return NULL;
    if (strcmp (stringval (name), "test") == 0)
	return function (test);
    return NULL;
}

main (int argc, char **argv)
{
    extern FILE * unref_debug;

    unref_debug = stderr;
    print_atom (expr_evaluate (string (argv[1]), test_cb));
    printf ("\n");
}
#endif
