#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <libgen.h>
#include <unistd.h>
#include "args.h"
#include "cache.h"
#include "cf.h"
#include "concat.h"
#include "dict.h"
#include "dirname.h"
#include "env.h"
#include "expr.h"
#include "global.h"
#include "netfind.h"
#include "pathnames.h"
#include "result.h"
#include "spawn.h"
#include "talloc.h"
#include "type.h"
#include "ugly.h"
#include "unpack.h"
#include "verbosity.h"

struct global global;

/*
 * attempt to download a particular library.  'lib' is of the form
 *
 * (libname url [candidate])
 */

int
try_download_one (struct result *x, Atom *lib)
{
    char *localfile;
    Atom *dict;
    char *libname;
    char *url;

    libname = stringval (this (lib));
    url = stringval (this (next (lib)));

#if 0
    printf ("try_download_one libname=%s uri=%s\n", libname, url);
#endif

    if (localfile = download_file_via_cache (url, libname)) {
	dict = unmeta ("candidate", next (next (lib)));
	dict = define (dict, "localfile", string (localfile));
#if 0
	fprint_dict (stdout, dict);
	printf ("\n");
	printf ("-> %s\n", localfile);
#endif
	return success (x);
    }
    else
	return failure (x, 1, "%s: %s download failed", libname, url);
}

/*
 * try to download a particular set of libraries.  'lib' is of the form
 *
 * ((libname1 url1 [candidate1])
 *  (libname2 url2 [candidate2])
 *  ...
 *  (libnameN urlN [candidateN])
 * )
 *
 * return 0 if all libs were downloaded
 */

int
try_download_libs (struct result *x, Atom *lib)
{
    while (!nullp (lib)) {
	if (try_download_one (x, this (lib)) != STATUS_OK)
	    return x->status;
	lib = next (lib);
    }
    return success (x);
}

/*
 * attempt to extract a particular library.  'lib' is of the form
 *
 * (libname url [candidate])
 */

int
try_extract_one (struct result *x, Atom *lib)
{
    FILE *fp;
    char *libname;		/* name of the library */
    Atom *metadata;		/* metadata for the library from index */
    char *metadata_fn;		/* filename of metadata file from package */
    char *package_fn;		/* local filename of package */
    char *temp_dirname;		/* temp dir where we unpack library */
    char *uri;			/* uri of library */

    libname = stringval (this (lib));
    uri = stringval (this (next (lib)));
    metadata = unmeta ("candidate", next (next (lib)));
    package_fn = stringval (nlookup (metadata, "localfile"));


    if (libname == NULL || uri == NULL || package_fn == NULL) {
	/* XXX print something useful here */
	return 0;
    }
    /*
     * HACK: if uri begins with '/' it's a locally-installed
     * lib (not a preloaded package) - and we don't need to extract it
     */
    if (*uri == '/')
	return success (x);

    /*
     * HACK: if uri begins with "file:///", it's a preloaded library
     * which is already unpacked into a local directory.  so we
     * don't need to download it.
     *
     * XXX maybe we should change this to localdir:/// or some such
     */
    if (strncmp (uri, "file:///", 8) == 0) {
	temp_dirname = uri + 8;
    }
    /*
     * otherwise, unpack the package unto a directory
     */
    else {
	temp_dirname = concat (getenv ("HOME"), "/NetBuild2/cache/",
			       hash_uri (uri), NULL);
	if (unpack_and_verify_package (x, temp_dirname, package_fn,
				       global.verbosity) != 0) {
	    /* XXX print something useful here */
	    return x->status;
	}
    }

    if ((metadata = read_metadata (metadata, temp_dirname)) == NULL)
	return failure (x, 1, "%s: cannot read metadata", libname);
    metadata = define (metadata, "localdir", string (temp_dirname));

    /*
     * HACK: if metadata doesn't include library.precedence, add a
     * dummy entry for it so that the sort doesn't fail
     */
    if (lookup (metadata, "library.precedence") == NULL)
	metadata = define (metadata, "library.precedence", string ("-9999"));

    return success (x);
}

/*
 * try to extract a particular set of libraries.  'lib' is of the form
 *
 * ((libname1 url1 [candidate1])
 *  (libname2 url2 [candidate2])
 *  ...
 *  (libnameN urlN [candidateN])
 * )
 *
 * return 0 if all libs were downloaded, else an error code
 */

int
try_extract_libs (struct result *x, Atom *lib)
{

#if 0
    printf ("trying to extract:\n");
    print_atom (lib);
    printf ("\n");
#endif

    while (!nullp (lib)) {
	if (try_extract_one (x, this (lib)) != 0)
	    return x->status;
	lib = next (lib);
    }

    return success (x);
}

/*
 * return nonzero iff the file named 'compiler' in directory 'dir'
 * is a symlink to something named "nbshim"
 */

static int
link_to_nbshim (Atom *dir, Atom *compiler)
{
    char buf[1024];
    char *ptr;
    char *fname;

    /*
     * empty string is treated as current directory
     */
    if (strcmp (stringval(dir), "") == 0)
	fname = stringval (compiler);
    else
	fname = concat (stringval (dir), "/", stringval (compiler), NULL);

    if (readlink (fname, buf, sizeof (buf)) == 0) {
	free (fname);

	/* see if link matches */
	if ((ptr = strrchr (buf, '/')) == NULL)
	    ptr = buf;
	if (strcmp (ptr, "nbshim") == 0)
	    return 1;
    }
    free (fname);
    return 0;
}

/*
 * copy path, removing any directories which either
 * a) match exclude_dir, or
 * b) $DIR/compiler is a symlink to something ending in nbshim
 */

static Atom *
fix_path (Atom *oldpath, Atom *compiler, char *exclude_dir)
{
    if (nullp (oldpath))
	return NULL;
    if (!stringp (this (oldpath)))
	return NULL;

    if (exclude_dir &&
	strcmp (stringval (this (oldpath)), exclude_dir) == 0) {
#if 0
	printf ("removing %s from PATH\n", stringval (this (oldpath)));
#endif
	return fix_path (next (oldpath), compiler, exclude_dir);
    }
    else if (link_to_nbshim (this (oldpath), compiler)) {
#if 0
	printf ("removing %s from PATH\n", stringval (this (oldpath)));
#endif
	return fix_path (next (oldpath), compiler, exclude_dir);
    }
    else
	return cons (this (oldpath),
		     fix_path (next (oldpath), compiler, exclude_dir));
}

/*
 * search 'path' (list of directories) for an executable named 'compiler'
 */

Atom *
find_compiler (Atom *compiler, Atom *path)
{
    char *filename;

    if (strchr (stringval(compiler), '/')) {
	/*
	 * 'compiler' is a full or relative path name
	 *
	 * if it's a link to nbshim, search the path for another
	 * compiler with the same base name.
	 * otherwise, just return the full path name
	 */
	Atom *bn = string (basename (stringval (compiler)));
	Atom *dn = string (my_dirname (stringval (compiler)));

	if (link_to_nbshim (dn, bn)) {
	    unref (dn);
	    compiler = bn;
	}
	else {
	    unref (bn);
	    unref (dn);
	    return compiler;
	}
    }
    if (nullp (path))
	return NULL;
    filename = concat (stringval (this (path)), "/", stringval (compiler),
		       NULL);
    if (access (filename, X_OK) == 0)
	return string (filename);
    return find_compiler (compiler, next (path));
}

/*
 * exec the command in 'arglist' 
 *
 * if the command matches global.argv0, modify the PATH environment
 * variable so that it doesn't contain our shim directory
 *
 * returns only if exec fails.
 */

int
exec_modified_path (Atom *arglist, Atom *envlist)
{
    char *bn = basename (global.argv0);
    Atom *compiler = NULL;
    Atom *pathlist = path2list (getenv ("PATH"));
    extern char **environ;
    Atom *newarglist;
    Atom *newpathlist;
    Atom *tmp;

    if (envlist == NULL)
	envlist = env2dict (environ);

    /*
     * set 'newpathlist' differently depending on how we were called
     *
     * a) if compiler matches global.argv0, remove shim directory from
     * newpathlist
     *
     * XXX need to make sure this agrees with the shim directory
     * installed by nb2 shell script
     *
     * XXX can there be more than one shim directory?
     *
     */
    if (strcmp (stringval (this (arglist)), global.argv0) == 0) {
	newpathlist = fix_path (pathlist, this (arglist), PATH_SHIM_DIR);
    }

    /*
     * b) if global.argv0 matches "nb" or "nb2" or "nbshim", _prepend_
     * shim directory to PATH (if it's not there already)
     */
    else if (strcmp (bn, "nb") == 0 ||
	     strcmp (bn, "nb2") == 0 ||
	     strcmp (bn, "nbshim") == 0) {
	if (strcmp (stringval (this (pathlist)), PATH_SHIM_DIR) != 0)
	    newpathlist = cons (string (PATH_SHIM_DIR), pathlist);
    }

    /*
     * c) if shim was called with a rooted path name, we still need
     * to remove it.
     */
    else if (strncmp (global.argv0, PATH_SHIM_DIR, strlen (PATH_SHIM_DIR)) == 0 &&
	global.argv0[strlen(PATH_SHIM_DIR)] == '/') {
	newpathlist = fix_path (pathlist, this (arglist), PATH_SHIM_DIR);
    }

    /*
     * d) otherwise, path remains unchanged
     *
     * XXX should this ever happen?
     */
    else {
	newpathlist = pathlist;
    }

    if ((compiler = find_compiler (this (arglist), newpathlist)) == NULL) {
	fprintf (stderr, "nb2: cannot find %s in $PATH\n",
		 stringval (this (arglist)));
	return 127;
    }
    newarglist = cons (compiler, next (arglist));

    envlist = define (envlist, 
		      "PATH", string (list2path (newpathlist)));

    /*
     * XXX modify environment to include requires
     */

    /*
     * now exec the subprocess with the modified environment
     *
     * XXX need to check verbosity level before printing
     */
    fprintf (stderr, ">>>");
    for (tmp = newarglist; !nullp (tmp); tmp = next (tmp))
	fprintf (stderr, " %s", stringval (this (tmp)));
    fprintf (stderr, "\n");

    return execve (stringval (compiler), list2argv (newarglist),
		   dict2env (envlist));
}

/*
 * given a list of mappings from libraries to packages, try each in
 * sequence until we find one that will link. 'liblist' is of the form
 * (
 * (("lapack" url1 attrs...) ("blas" url2 attrs...))
 * (("lapack" url1 attrs...) ("blas" url3 attrs...))
 * )
 */

void
print_libset (Atom *libset)
{
    Atom *ptr;
    Atom *lib;
    for (ptr = libset; !nullp (ptr); ptr = next (ptr)) {
	lib = this (ptr);
	printf ("%s %s\n", stringval (this (lib)),
		stringval (this (next (lib))));
    }
}

int
do_link (Atom *arglist, Atom *liblist)
{
    struct result x = { 0 };

    if (nullp (liblist)) {
	Atom *child_args = make_child_arglist (arglist, NULL);

	return exec_modified_path (child_args, NULL);
    }
    while (!nullp (liblist)) {
	if (VERBOSE_ENOUGH (VERBOSITY_LIB_SELECTION)) {
	    printf ("nb2: trying libraries:\n");
	    print_libset (this (liblist));
	}
	if (try_download_libs (&x, this (liblist)) == STATUS_OK) {
	    /* download successful.  now extract */

	    if (try_extract_libs (&x, this (liblist)) == STATUS_OK) {
		/*
		 * extract successful. now link
		 */
		extern char **environ;
		Atom *child_args = make_child_arglist (arglist,
						       this (liblist));
		Atom *child_env = env2dict (environ);
		Atom *selection = this (liblist);
		Atom *ptr;

		/*
		 * generate environment variables of the form
		 * NB2_LIB_{libname}=file:///{localdir}/
		 * for any pre-selected libraries
		 */
		for (ptr = selection; !nullp (ptr); ptr=next(ptr)) {
		    Atom *lib = this (ptr);
		    char *name;
		    char *url;
			
		    name = concat ("NB2_LIB_",
				   stringval (ugly_encode (this (lib))),
				   NULL);
		    url = concat ("file:///",
				  stringval (nlookup (unmeta ("candidate",
							      next (next (lib))),
						      "localdir")),
				  "/", NULL);
		    child_env =	define (child_env, name, string (url));
		    free (name);
		    free (url);
		}
		/*
		 * this assumes that if the link command fails, we
		 * want to fail also.  we could have an option to
		 * retry the link with some other set of libraries,
		 * in which case we'd need to fork a subprocess.
		 */
		return exec_modified_path (child_args, child_env);
	    }
	    else if (VERBOSE_ENOUGH (VERBOSITY_LIB_SELECTION)) {
		fprintf (stderr, "nb2: %s\n", x.msg);
	    }
	}
	else if (VERBOSE_ENOUGH (VERBOSITY_LIB_SELECTION)) {
	    fprintf (stderr, "nb2: %s\n", x.msg);
	}
	/*
	 * some sort of download or extract failure.
	 *
	 * XXX print something?
	 */
	liblist = next (liblist);
    }
    fprintf (stderr,
	     "nb2: unable to download and extract required libraries\n");
    return -1;
}

/*
 * make sure required files are in place
 */

int
sanity ()
{
    /*
     * NB: these files should always be relative to user's home area,
     * so that the individual user can decide whom he trusts by
     * altering the GPG keyrings and trust parameters.
     */
    static char *files_to_check[] = {
	"/gnupg/pubring.gpg",
	"/gnupg/secring.gpg",
	"/gnupg/trustdb.gpg",
	NULL,
    };
    int i;

    for (i = 0; files_to_check[i] != NULL; ++i) {
	char *foo = concat (global.path_prefix,
			    files_to_check[i], NULL);

	if (access (foo, R_OK) < 0) {
	    fprintf (stderr, "cannot access required file %s\n", foo);
	    fprintf (stderr, "run 'nb2 -setup' to install required files\n");
	    exit (1);
	}
	free (foo);
    }
}

void
usage (char *prefix)
{
    static char *commands[] = {
	"[ options ] COMMAND",
	"[ options ] -download URL FILENAME",
	"[ options ] -dump FILENAME",
	"[ options ] -e EXPRESSION",
	"[ options ] -md5 FILENAME...",
	"[ options ] -setup",
	NULL,
    };
    static char *options[] = {
	"-platform FILENAME",
	"-r[equire] LIBNAME",
	"-share MODE",
	"-share DIRECTORY",
	"-verbosity FILENAME",
	"DIRECTORY must contain a slash",
	"MODE is one of local, system, script, netload",
	NULL,
    };
    int i;

    fprintf (stderr, "usage:\n");
    for (i = 0; commands[i]; ++i)
	fprintf (stderr, "    %s %s\n", prefix, commands[i]);
    fprintf (stderr, "options:\n");
    for (i = 0; options[i]; ++i)
	fprintf (stderr, "    %s\n", options[i]);
    exit (1);
}

/*
 * scan the environment looking for variables of the form
 * NB2_LIB_libname=file:///directory
 * and for each one found, add an entry to liblist that
 * is like the entry that would be created by process_args()
 * if that library were mentioned on the command-line.
 * this will cause us to add those librarys' compile args
 * whenever we run a compiler that accepts them.
 */

Atom *
add_env_libs (Atom *liblist)
{
    extern char **environ;
    int i;
    Atom *result = liblist;

    for (i = 0; environ[i]; ++i) {
	if (strncmp (environ[i], "NB2_LIB_", 8) == 0) {
	    char *equal = strchr (environ[i]+8, '=');
	    Atom *libname;
	    Atom *ptr;
	    Atom *uri;
	    Atom *localdir;

	    if (equal == NULL || equal[1] == '\0')
		continue;
	    libname = ugly_decode (stringn (environ[i]+8,
					    equal-(environ[i]+8)));
	    uri = string (equal + 1);
	    if (strncmp (equal + 1, "file:///", 8) != 0)
		continue;
	    localdir = string (equal + 9);
	    
	    if (ptr = lookup (result, stringval (libname))) {
		/* already in liblist */
		;
	    }
	    else
		result = define (result, stringval(libname),
				 cons (cons (string (":preload"),
					     integer (1)),
				       NULL));
	}
    }
    return result;
}

/*
 * parse the command-line options
 */

void
process_options (int *argcp, char ***argvp, char *usage_prefix)
{
    int argc = *argcp;
    char **argv = *argvp;
    int i;

    /*
     * default some options from environment variables
     */
    if (getenv ("NB2_VERBOSITY"))
	global.verbosity = atoi (getenv ("NB2_VERBOSITY"));

    while (argv[0] != NULL) {
	if (strcmp (argv[0], "-platform") == 0) {
	    if (argv[1] == NULL)
		usage (usage_prefix);
	    global.platform_filename = argv[1];
	    argv += 2;
	    argc -= 2;
	}
	else if (strcmp (argv[0], "-require") == 0 || strcmp (argv[0], "-r") == 0) {
	    if (argv[1] == NULL)
		usage (usage_prefix);
	    global.required_libs =
		cons (cons (string (argv[1]),
			    cons (cons (string (":required"), integer (1)),
				  NULL)),
		      global.required_libs);
	    argv += 2;
	    argc -= 2;
	}
	else if (strcmp (argv[0], "-share") == 0) {
	    if (argv[1] == NULL)
		usage (usage_prefix);
	    if (strcmp (argv[1], "local") == 0)
		global.shared_lib_mode = SHARED_LOCAL;
	    else if (strcmp (argv[1], "system") == 0)
		global.shared_lib_mode = SHARED_SYSTEM;
	    else if (strcmp (argv[1], "script") == 0)
		global.shared_lib_mode = SHARED_SCRIPT;
	    else if (strcmp (argv[1], "netload") == 0)
		global.shared_lib_mode = SHARED_NETLOAD;
	    else if (strchr (argv[1], '/')) {
		char *ptr;

		/* delete trailing slash */
		if (ptr != argv[1] && (ptr = strchr (argv[1], '/'))[1] == '\0')
		    *ptr = '\0';

		global.shared_lib_mode = SHARED_USER_SPECIFIED_DIR;
		global.shared_dir = strdup (argv[1]);
	    }
	    else
		usage (usage_prefix);
	    argv += 2;
	    argc -= 2;
	}
	else if (strcmp (argv[0], "-verbosity") == 0) {
	    if (argv[1] == NULL)
		usage (usage_prefix);
	    global.verbosity = atoi (argv[1]);
	    argv += 2;
	    argc -= 2;
	}
	/*
	 * note that other arguments beginning with - are 
	 * handled by the caller - they are really commands
	 * not options
	 */
	else
	    goto done;
    }

 done:
    *argcp = argc;
    *argvp = argv;
    return;
}

/*
 * handle an external command (i.e. one not handled by nbshim)
 */

void
external_command (int argc, char **argv, char *usage_prefix)
{
    Atom *root;
    /* Atom *argpats; */
    Atom *path;
    Atom *liblist;

    process_options (&argc, &argv, usage_prefix);
    if (global.verbosity > 1) {
	fprintf (stderr, "NetBuild v2.0 (built %s)\n", BUILD_DATE);
	fprintf (stderr, "verbosity = %d\n", global.verbosity);
    }
    platform_init (global.platform_filename);
    root = find_config ("default.config", argv[0]);
    /* argpats = lookup (root, ":args"); */

    /*
     * XXX what to do with Library-Path and $LD_LIBRARY_PATH?
     * need to end up so that global.libpath contains everything,
     * ideally in the right order.  
     *
     * XXX also, we probably need a way to do platform-specific
     * or compiler-specific initializations - such as different
     * platforms use different environment variables for the
     * default library search path.
     */

    if (path = nlookup (root, "Library-Path"))
	global.libpath = path2list (stringval (path));
    else
	global.libpath = path2list ("/lib:/usr/lib");

    process_args (argv, root);
#if 0
    printf ("argstuff.libs=\n");
    print (argstuff.libs);
    printf ("global.required_libs=\n");
    print (global.required_libs);
#endif

    argstuff.libs = add_env_libs (argstuff.libs);
    liblist = find_libs (argstuff.libs, global.libpath);

    sanity ();
    exit (do_link (argstuff.arglist, liblist));
}

/*
 * handle an internal command (one handled by nbshim)
 */

void
internal_command (int argc, char **argv, char *usage_prefix)
{
    process_options (&argc, &argv, usage_prefix);
    if (global.verbosity > 1) {
	fprintf (stderr, "NetBuild v2.0 (built %s)\n", BUILD_DATE);
	fprintf (stderr, "verbosity = %d\n", global.verbosity);
    }
    platform_init (global.platform_filename);

    if (argv[0] == NULL)
	usage (usage_prefix);
    else if (strcmp (argv[0], "-download") == 0)
	nb_download (argv[1], argv[2]);
    else if (strcmp (argv[0], "-dump") == 0)
	nb_dump (argc - 1, argv + 1);
    else if (strcmp (argv[0], "-e") == 0) {
	nb_expr (argv + 1);
    }
    else if (strcmp (argv[0], "-md5") == 0) {
	int i;
	for (i = 1; argv[i]; ++i)
	    nb_md5 (argv[i]);
    }
    else if (strcmp (argv[0], "-setup") == 0)
	nb_setup ();	
    else {
	/* XXX not sure if this is what we should do */
	external_command (argc, argv, usage_prefix);
    }
    exit (0);  
}

/*
 * try to decide whether this is an internal or external
 * command
 */

void
dispatch (int argc, char **argv, char *usage_prefix)
{
    if (*argv[0] == '-')
	usage (usage_prefix);
    else if (strcmp (argv[0], "nb") == 0 || strcmp (argv[0], "nb2") == 0)
	internal_command (argc-1, argv+1, usage_prefix);
    else
	external_command (argc, argv, usage_prefix);
}

/*
 * look for an executable with name 'basename' in 'pathlist
 *
 * XXX compare with find_compiler() above
 */

Atom *
find_executable (Atom *basename, Atom *pathlist)
{
    Atom *ptr;

    if (*(stringval (basename)) == '/')
	return basename;
    for (ptr = pathlist; !nullp (ptr); ptr=next(ptr)) {
	char *pathname = concat (stringval (this (ptr)), "/",
				 stringval (basename), NULL);
	if (access (pathname, X_OK) == 0)
	    return string (pathname);
	free (pathname);
    }
    return NULL;
}

/*
 * see if we are supposed to run ourselves under the debugger
 */

void
check_debugger_trap (int argc, char **argv)
{
    int i;
    char *debug_trap = getenv ("NB2_DEBUG_TRAP");
    Atom *debugger;
    Atom *pathlist = path2list (getenv ("PATH"));
    Atom *executable;
    extern char **environ;
    Atom *envlist = env2dict (environ);

    if (debug_trap == NULL)
	return;
    if (getenv ("NB2_DEBUGGER") == NULL)
	debugger = string ("gdb");
    else
	debugger = string (getenv ("NB2_DEBUGGER"));

    for (i = 0; i < argc; ++i)
	if (strcmp (debug_trap, argv[i]) == 0)
	    break;
    if (i >= argc)
	return;

    debugger = find_executable (debugger, pathlist);
    if (*argv[0] == '/')
	executable = string (argv[0]);
    else
	executable = find_executable (string (argv[0]), pathlist);


    envlist = define (envlist, "NB2_DEBUG_TRAP", NULL);

    fflush (stdout);
    fflush (stderr);
    if (!isatty(0) || !isatty(1)) {
	freopen ("/dev/tty", "r", stdin);
	freopen ("/dev/tty", "w", stdout);
	freopen ("/dev/tty", "w", stderr);
    }
    printf ("\n\n*** nb2 debugger trap - command is\n");
    for (i = 0; i < argc; ++i)
	printf ("%s%c", argv[i], i == argc-1 ? '\n' : ' ');

    if (executable == NULL) {
	fprintf (stderr, "could not find debugger %s\n", stringval (debugger));
	exit (1);
    }


    execve (stringval (debugger),
	    list2argv (list (debugger, executable, NULL)),
	    dict2env (envlist));

    fprintf (stderr, "execve failed: %s\n", strerror (errno));
    fprintf (stderr, "pathname = %s\n", stringval (debugger));
    fprintf (stderr, "arglist = ");
    fprint_atom (stderr, list (debugger, executable, NULL));
    fprintf (stderr, "\n");
    exit (1);
}

main (int argc, char **argv)
{
    char *bn;

    check_debugger_trap (argc, argv);

    global_init (argv[0]);

    /*
     * if we were called with the name nbshim, act as if we were
     * called as argv[1] and argv[2..n] are the arguments.
     * this is mostly for debugging, as it's difficult to set
     * argv[0] from the debugger.
     */

    bn = basename (argv[0]);
    if (strcmp (bn, "nbshim") == 0) {
	if (argc <= 1)
	    usage (bn);
	else
	    dispatch (argc-1, argv+1, bn);
    }
    else {
	argv[0] = bn;
	dispatch (argc, argv, argv[0]);
    }
}
