#include <stdio.h>
#include <stdarg.h>
#include "talloc.h"
#include "type.h"

FILE *unref_debug;

Atom *
ref (Atom *a)
{
    if (a == NULL)
	return NULL;
    a->refcount++;
    return a;
}

void
unref (Atom *a)
{
    if (a == NULL)
	return;
    if (a->type == freedType) {
	if (unref_debug)
	    fprintf (unref_debug, "freeing already free'd node\n");
	return;
    }
    if (--(a->refcount) <= 0) {
	switch (a->type) {
	case nullType:
	case integerType:
	    if (unref_debug)
		fprintf (unref_debug, "freeing integer %d\n", intval (a));
	    break;
	case stringType:
	    if (unref_debug)
		fprintf (unref_debug, "freeing string %s\n", stringval (a));
	    break;
	case errorType:
	    if (unref_debug)
		fprintf (unref_debug, "freeing error node\n");
	    break;
	case functionType:
	    if (unref_debug)
		fprintf (unref_debug, "freeing function node\n");
	    break;
	default:
	    break;
	case listType:
	    if (unref_debug)
		fprintf (unref_debug, "freeing list node\n");
	    unref (((List *)a)->this);
	    unref (((List *)a)->next);
	    ((List *)a)->next = NULL;
	    ((List *)a)->this = NULL;
	    break;
	}
	a->type = freedType;
	free (a);
    }
}

/* constructors */

Atom *
integer (int i)
{
    Integer *result;

    result = TALLOC (1, Integer);
    result->atom.refcount = 0;
    result->atom.type = integerType;
    result->val = i;
    return (Atom *) result;
}

Atom *
string (char *s)
{
    String *result;
    int len;

    if (s == NULL)
	s = "";
    len = strlen (s);

    result = (String *) xalloc (sizeof (String) + len + 1);
    result->atom.refcount = 0;
    result->atom.type = stringType;
    strcpy (result->str, s);
    return (Atom *) result;
}

Atom *
stringn (char *s, int n)
{
    String *result;
    int len;

    if (s == NULL)
	s = "";
    len = strlen (s);
    if (len > n)
	len = n;

    result = (String *) xalloc (sizeof (String) + len + 1);
    result->atom.refcount = 0;
    result->atom.type = stringType;
    strncpy (result->str, s, len);
    result->str[len] = '\0';
    return (Atom *) result;
}

Atom *
function (Atom *(*func)())
{
    Function *result;

    result = TALLOC (1, Function);
    result->atom.refcount = 0;
    result->atom.type = functionType;
    result->func = func;
    return (Atom *) result;
}

Atom *
empty_list ()
{
    List *result;

    result = TALLOC (1, List);
    result->atom.refcount = 0;
    result->atom.type = listType;
    result->this = result->next = NULL;
    return (Atom *) result;
}

Atom *
list (Atom *arg0, ...)
{
    List *result;
    List *last;
    Atom *arg;
    va_list ap;

    if (arg0 == NULL)
	return NULL;

    result = (List *) empty_list ();
    result->this = ref (arg0);
    last = result;

    va_start (ap, arg0);
    while (arg = va_arg (ap, Atom *)) {
	List *n = (List *) empty_list ();
	n->this = ref (arg);
	last->next = ref ((Atom *) n);
	last = n;
    }
    va_end (ap);
    return (Atom *) result;
}

Atom *
list2 (Atom *arg0, Atom *arg1)
{
    return list (arg0, arg1, NULL);
}

Atom *
meta (char *typename, Atom *arg)
{
    Meta *result;
    int len;

    if (typename == NULL)
	return NULL;
    len = strlen (typename);

    result = (Meta *) xalloc (sizeof (Meta) + len + 1);
    result->atom.refcount = 0;
    result->atom.type = metaType;
    strcpy (result->typename, typename);
    result->next = ref (arg);
    return (Atom *) result;
}

Atom *
unmeta (char *typename, Atom *arg)
{
    if (!metap (arg))
	return error ();
    if (strcmp (((Meta *) arg)->typename, typename) != 0)
	return error ();
    return ((Meta *)arg)->next;
}

char *
typename (Atom *arg)
{
    if (metap (arg))
	return ((Meta *)arg)->typename;
    switch (arg->type) {
    case nullType: return "[null]";
    case integerType: return "[integer]";
    case stringType: return "[string]";
    case functionType: return "[function]";
    case listType: return "[list]";
    case errorType: return "[error]";
    case freedType: return "**FREED**";
    default: return "**UNKNOWN**";
    }
}

Atom *
error ()
{
    Error *result;

    result = TALLOC (1, Error);
    result->atom.refcount = 0;
    result->atom.type = errorType;
    return (Atom *) result;
}

int
intval (Atom *a)
{
    if (a == NULL)
	return 0;
    return ((Integer *) a)->val;
}

/* string operators */

char *
stringval (Atom *a)
{
    if (a == NULL)
	return NULL;
    return ((String *) a)->str;
}

Atom *
string_concat (Atom *a, ...)
{
    String *result;
    va_list ap;
    int length = 0;
    Atom *arg;

    if (!stringp (a))
	return NULL;
    length = strlen (stringval (a));
    va_start (ap, a);
    while (arg = va_arg (ap, Atom *)) {
	if (!stringp (arg))
	    return NULL;
	length += strlen (stringval (arg));
    }
    va_end (ap);
    result = (String *) xalloc (sizeof (String) + length + 1);
    result->atom.refcount = 0;
    result->atom.type = stringType;
    strcpy (result->str, stringval (a));
    va_start (ap, a);
    while (arg = va_arg (ap, Atom *))
	strcat (result->str, stringval (arg));
    va_end (ap);
    return (Atom *) result;
}

/* list operators */

Atom *
this (Atom *a)
{
    if (a == NULL)
	return NULL;
    if (a->type != listType)
	return NULL;
    return (((List *)a)->this);
}

Atom *
next (Atom *a)
{
    if (a == NULL)
	return NULL;
    if (a->type != listType)
	return NULL;
    return (((List *)a)->next);
}

void
set_this (Atom *a, Atom *b)
{
    List *l = (List *) a;

    if (a == NULL)
	return;
    if (a->type != listType)
	return;
    if (l->this != b) {
	unref (l->this);
	l->this = ref (b);
    }
}

void
set_next (Atom *a, Atom *b)
{
    List *l = (List *) a;

    if (a == NULL)
	return;
    if (a->type != listType)
	return;
    if (l->next != b) {
	unref (l->next);
	l->next = ref (b);
    }
}

Atom *
cons (Atom *arg0, Atom *arg1)
{
    List *result = (List *) empty_list ();

    result->this = ref (arg0);
    result->next = ref (arg1);
    return (Atom *) result;
}

Atom *
append (Atom *arg0, Atom *arg1)
{
    Atom *result;
    Atom *ptr;

    if (arg0 == NULL)
	return arg1;
    if (arg0->type != listType)
	return NULL;
    result = arg0;
    for (ptr = arg0; listp(ptr) && next(ptr); ptr=next(ptr));
    set_next (ptr, arg1);
    return result;
}

/* type predicates */

int
nullp (Atom *a)
{
    return a == NULL;
}

int
integerp (Atom *a)
{
    if (a == NULL)
	return 0;
    return (a->type == integerType);
}

int 
listp (Atom *a)
{
    if (a == NULL)
	return 0;
    return (a->type == listType);
}

int 
stringp (Atom *a)
{
    if (a == NULL)
	return 0;
    return (a->type == stringType);
}

int
metap (Atom *a)
{
    if (a == NULL)
	return 0;
    return (a->type == metaType);
}

/* utilties */

void fprint_atom ();

static void
fprint_list (FILE *fp, List *a, int flags, int level)
{
    int dflag = flags & 01;

    fprint_atom_generic (fp, a->this, flags, level);
    if (a->next == NULL)
	;
    else if (listp (a->next)) {
	fprintf (fp, " ");
	fprint_list (fp, (List *) a->next, flags, level);
    }
    else {
	fprintf (fp, " ");
	fprint_atom_generic (fp, a->next, flags, level);
    }
}

void
fprint_atom_generic (FILE *fp, Atom *a, int flags, int depth)
{
    Atom *p;
    int dflag = flags & 01;
    int iflag = flags & 02;
    int i;

    if (a == NULL) {
	fprintf (fp, "nil");
	return;
    }
    if (dflag)
	fprintf (fp, "{%d} ", a->refcount);
    switch (a->type) {
    case integerType:
	fprintf (fp, "%d", ((Integer *)a)->val);
	break;
    case stringType:
	fprintf (fp, "\"%s\"", ((String *)a)->str);
	break;
    case functionType:
	fprintf (fp, "%x()", ((Function *)a)->func);
	break;
    case listType:
	fprintf (fp, "(");
	fprint_list (fp, (List *) a, flags, depth);
	fprintf (fp, ")");
	break;
    case errorType:
	fprintf (fp, "error");
	break;
    case metaType:
	fprintf (fp, "[%s]", ((Meta *)a)->typename);
	if (depth > 0) {
	    fprint_atom_generic (fp, ((Meta *)a)->next, flags, depth-1);
	    fprintf (fp, "\n");
	}
	break;
    case freedType:
	fprintf (fp, "**FREED NODE**");
	break;
    }
}

void
fprint_atom (FILE *fp, Atom *a)
{
    fprint_atom_generic (fp, a, 0, 0);
}

void
print_atom (Atom *a)
{
    fprint_atom_generic (stdout, a, 02, 0);
}

/*
 * this is **immensely** useful for whenever you want to print an atom
 * from within the debugger
 */

void
print (Atom *a)
{
    FILE *fp = fopen ("/dev/tty", "w");
    fprint_atom_generic (fp, a, 02, 0);
    fprintf (fp, "\n");
    fclose (fp);
}

void
nprint (Atom *a, int depth)
{
    FILE *fp = fopen ("/dev/tty", "w");
    fprint_atom_generic (fp, a, 02, depth);
    fprintf (fp, "\n");
    fclose (fp);
}

#ifdef TEST
main ()
{
    Atom *a;
    a = NULL;
    fprint_atom (stdout, a);
    printf ("\n");

    a = empty_list();
    fprint_atom (stdout, a);
    printf ("\n");

    a = string ("one");
    fprint_atom (stdout, a);
    printf ("\n");

    a = list (string ("one"), NULL);
    fprint_atom (stdout, a);
    printf ("\n");

    a = list (string ("one"), string ("two"), NULL);
    fprint_atom (stdout, a);
    printf ("\n");

    a = list (string ("one"), string ("two"), string ("three"), NULL);
    fprint_atom (stdout, a);
    printf ("\n");

    a = list (string ("one"), string ("two"), string ("three"), string ("four"), NULL);
    fprint_atom (stdout, a);
    printf ("\n");

    a = list (list (string ("one"), string ("1"), NULL),
	      list (string ("two"), string ("2"), NULL),
	      NULL);
    fprint_atom (stdout, a);
    printf ("\n");

    a = list (list (string ("one"), string ("1"), NULL),
	      list (string ("two"), string ("2"), NULL),
	      list (string ("three"), string ("3"), NULL),
	      NULL);
    fprint_atom (stdout, a);
    printf ("\n");

    a = list (list (string ("four"), string ("4"), NULL), a, NULL);
    fprint_atom (stdout, a);
    printf ("\n");

    a = list (string ("one"), integer (1), NULL);
    a = list (list (string ("two"), integer (2), NULL), a, NULL);
    a = list (list (string ("Three"), integer (3), NULL), a, NULL);
    fprint_atom (stdout, a);
    printf ("\n");

    a = NULL;
    a = cons (list2 (string ("one"), integer (1)), a);
    fprint_atom (stdout, a);
    printf ("\n");
    a = cons (list2 (string ("two"), integer (2)), a);
    fprint_atom (stdout, a);
    printf ("\n");
    a = cons (list2 (string ("three"), integer (3)), a);
    fprint_atom (stdout, a);
    printf ("\n");
    a = meta ("foo", a);
    fprint_atom (stdout, a);
    printf ("\n");
    fprint_atom (stdout, unmeta ("foo", a));
    printf ("\n");
    fprint_atom (stdout, unmeta ("bar", a));
    printf ("\n");
    printf ("%s\n", typename (a));
}
#endif
