#ifndef TYPE_H
#define TYPE_H

#include <stdio.h>

/*
 * data types for nb2
 * these data types are used in config files,
 * constraint expressions, and internal data structures
 *
 * primitive data types (atoms) are integer, string, function, and list.
 *
 * lists are linked-lists of atoms.  
 *
 * functions take an atom as input, and return an atom as result, but
 * since any atom can be a list, all functions potentially have
 * variadic input/output
 *
 * For simplicity and robustness, an entire atom (including its name
 * and/or contents of the string) is malloc'ed or free'd in a single
 * call.
 *
 * This was obviously inspired by lisp, but don't blame me that it's
 * not really much like lisp.  I like the generality of the lisp data
 * structures but putting a real lisp interpreter into nb would be
 * overkill.  And nb wants to be portable so I don't want it to have
 * to depend on someone else's lisp interpreter either.  (similarly
 * for other languages that also support on-the-fly construction
 * of complex data structures).
 */

enum SymType {
    nullType,
    integerType,
    stringType,
    functionType,
    listType,
    errorType,
    metaType,
    freedType
};

typedef struct atom {
    enum SymType type;
    int refcount;
} Atom;

typedef struct integer {
    struct atom atom;
    int val;
} Integer;

typedef struct string {
    struct atom atom;
    int len;
    char str[1];
} String;

typedef struct function {
    struct atom atom;
    struct atom *(*func)();
} Function;

typedef struct list {
    struct atom atom;
    struct atom *this;
    struct atom *next;
} List;

typedef struct meta {
    struct atom atom;
    struct atom *next;
    int len;
    char typename[1];
} Meta;

typedef struct error {
    struct atom atom;
} Error;

Atom *ref (Atom *);
void unref (Atom *);

/* constructors */
Atom *integer (int);
Atom *string (char *);
Atom *stringn (char *, int);
Atom *function (Atom *(*)());
Atom *empty_list ();
Atom *list (Atom *, ...);
Atom *list2 (Atom *, Atom *);
Atom *error ();

/* integer operator */
int intval (Atom *);

/* string operators */
char *stringval (Atom *);
Atom *string_concat (Atom *, ...);

/* list operators */
Atom *this (Atom *);
Atom *next (Atom *);
void set_this (Atom *, Atom *);
void set_next (Atom *, Atom *);
Atom *cons (Atom *, Atom *);
Atom *append (Atom *, Atom *);

/* meta operators */
Atom *meta (char *typename, Atom *);
Atom *unmeta (char *typename, Atom *);
char *typename (Atom *);

/* type predicates */
int nullp (Atom *);
int integerp (Atom *);
int listp (Atom *);
int stringp (Atom *);
int metap (Atom *);
void fprint_atom_generic (FILE *, Atom *, int, int);
void fprint_atom (FILE *, Atom *);
void print_atom (Atom *);

#endif /* TYPE_H */
